/* the main window 
 *
 * $Id: g2main.h,v 1.1.1.1 2004/06/23 02:01:44 miq Exp $
 *
 */

#ifndef G2_MAIN_WINDOW_H
#define G2_MAIN_WINDOW_H

#include <gtk/gtk.h>

typedef struct {
    gint id;
    GtkWidget *win;
} g2Win;

GSList* keyBuffer;
gint clickX;
gint clickY;
gint clickMod;

void g2_init_main_window(int *argcp, char **argv);
GtkWidget *g2_get_main_window();
GtkWidget *g2_get_equipment_window();
gint g2_create_window(gint type);
void g2_exit_windows(const gchar* string);

#endif                          /* G2_MAIN_WINDOW_H */
